use std::{
    collections::VecDeque,
    fmt::Debug,
    marker::PhantomData,
    mem::size_of,
    net::SocketAddr,
    num::{Saturating, Wrapping},
    time::{Duration, Instant},
};

use futures::{future::Fuse, Future, FutureExt};
use tokio::{
    io::{AsyncReadExt, AsyncWriteExt},
    net::{TcpListener, TcpStream},
    select,
    sync::mpsc::{error::TryRecvError, Receiver, Sender},
};
use zerocopy::{AsBytes, FromBytes, FromZeroes};

const PING_TIMEOUT: Duration = Duration::from_secs(5);
const PING_INTERVAL: Duration = Duration::from_secs(1);

#[derive(Debug, serde::Deserialize, serde::Serialize)]
enum Message<T> {
    Payload(T),
    Connect { reconnect: bool, last_id: u64 },
    Nack,
    Ack { id: u64 },
    Ping,
}

#[derive(AsBytes, FromBytes, FromZeroes, Clone, Copy)]
#[repr(transparent)]
struct SegmentSize(zerocopy::network_endian::U32);

impl Debug for SegmentSize {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        f.debug_struct("SegmentSize")
            .field("size", &self.size())
            .field("complete", &!self.is_incomplete())
            .finish()
    }
}

const U32_HIGH_BIT: u32 = 1 << (u32::BITS - 1);
const U31_MAX: u32 = U32_HIGH_BIT - 1;

impl SegmentSize {
    fn from_total(total_size: u64) -> Self {
        // use the highest bit to mark if a packet is incomplete
        let size_field = if total_size <= U31_MAX as u64 {
            total_size as u32
        } else {
            // partial segment.
            // set partial bit and use maximum size
            // U32_HIGH_BIT | U31_MASK
            u32::MAX
        };

        Self(size_field.into())
    }
    fn is_incomplete(&self) -> bool {
        self.0.get() & U32_HIGH_BIT != 0
    }
    fn size(&self) -> u32 {
        self.0.get() & U31_MAX
    }
}

struct SendingState {
    buffer: Vec<u8>,
    i: usize,
    segment_start: usize,
}

struct StreamWriteHalf<T, S> {
    stream: S,
    sending_state: Option<SendingState>,
    _phantom: PhantomData<fn(T) -> T>,
}

impl<T, S> StreamWriteHalf<T, S>
where
    T: serde::Serialize,
    S: AsyncWriteExt + Unpin,
{
    fn new(stream: S) -> Self {
        Self {
            stream,
            sending_state: None,
            _phantom: PhantomData,
        }
    }

    // Cancellation safe
    async fn finish_send(&mut self) -> Option<()> {
        if let Some(state) = &mut self.sending_state {
            loop {
                let segment_size: SegmentSize = SegmentSize::read_from(
                    &state.buffer
                        [state.segment_start..state.segment_start + size_of::<SegmentSize>()],
                )
                .unwrap();

                let segment_end =
                    state.segment_start + size_of::<SegmentSize>() + segment_size.size() as usize;

                while state.i < segment_end {
                    let data = &state.buffer[state.i..segment_end];

                    let n = self.stream.write(data).await.ok()?;
                    if n == 0 {
                        // remote closed the connection
                        return None;
                    }
                    state.i += n;
                }

                if segment_size.is_incomplete() {
                    assert_eq!(state.i, segment_end);
                    state.segment_start = segment_end;

                    let segment_buffer = &mut state.buffer[state.segment_start..];
                    let segment_size = SegmentSize::from_total(
                        segment_buffer
                            .len()
                            .checked_sub(size_of::<SegmentSize>())
                            .unwrap() as u64,
                    );
                    segment_buffer[0..size_of::<SegmentSize>()]
                        .copy_from_slice(segment_size.as_bytes());
                } else {
                    break;
                }
            }
        }

        self.sending_state = None;

        Some(())
    }

    fn start_sending(&mut self, message: &Message<T>) {
        assert!(
            self.sending_state.is_none(),
            "tried to queue a message before finishing an old message"
        );

        let message_bytes = bincode::serialized_size(message).unwrap();
        let buffer_size: usize = (size_of::<SegmentSize>() as u64)
            .checked_add(message_bytes.min(U31_MAX as u64))
            .expect("message too large")
            .try_into()
            .expect("message too large for platform");

        let mut buffer = vec![0; buffer_size];

        // TODO: Writer?
        // write serialized message into buffer offset by the size of one u32 to leave space for the size field.
        bincode::serialize_into(
            &mut std::io::Cursor::new(&mut buffer[size_of::<SegmentSize>()..]),
            message,
        )
        .unwrap();

        let segment_size = SegmentSize::from_total(message_bytes);

        // write size + fragmentation info in front of message buffer
        buffer[0..size_of::<SegmentSize>()].copy_from_slice(segment_size.as_bytes());

        self.sending_state = Some(SendingState {
            buffer,
            i: 0,
            segment_start: 0,
        });
    }

    fn has_partial_send(&self) -> bool {
        self.sending_state.is_some()
    }

    // Not cancellation safe. Call finish_send until done before calling send again if a send was interupted
    async fn send<'a>(&mut self, message: &'a Message<T>) -> Option<()> {
        self.start_sending(message);
        self.finish_send().await
    }
}

struct ReceivingState {
    buffer: Vec<u8>,
    i: usize,
    segment_size: SegmentSize,
}

struct HeaderRecvState {
    segment_size: SegmentSize,
    i: usize,
}

struct StreamReadHalf<T, S> {
    stream: S,
    receiving_state: Option<ReceivingState>,
    header_state: Option<HeaderRecvState>,
    _phantom: PhantomData<fn(T) -> T>,
}

impl<T, S> StreamReadHalf<T, S>
where
    T: for<'de> serde::Deserialize<'de>,
    S: AsyncReadExt + Unpin,
{
    fn new(stream: S) -> Self {
        Self {
            stream,
            receiving_state: None,
            header_state: None,
            _phantom: PhantomData,
        }
    }

    // cancellation safe
    async fn recv_header(
        header_state: &mut Option<HeaderRecvState>,
        stream: &mut S,
    ) -> Option<SegmentSize> {
        if header_state.is_none() {
            *header_state = Some(HeaderRecvState {
                segment_size: SegmentSize::new_zeroed(),
                i: 0,
            });
        }
        let header = header_state.as_mut().unwrap();

        while header.i < size_of::<SegmentSize>() {
            let n = stream
                .read(&mut header.segment_size.as_bytes_mut()[header.i..])
                .await
                .ok()?;

            if n == 0 {
                return None;
            }

            header.i += n;
        }

        let segment_size = header.segment_size;
        *header_state = None;
        Some(segment_size)
    }

    // TODO: make cancellation safe
    async fn recv(&mut self) -> Option<Message<T>> {
        let state = if let Some(state) = self.receiving_state.as_mut() {
            state
        } else {
            let segment_size = Self::recv_header(&mut self.header_state, &mut self.stream).await?;
            self.receiving_state = Some(ReceivingState {
                buffer: vec![0; segment_size.size() as usize],
                i: 0,
                segment_size,
            });
            self.receiving_state.as_mut().unwrap()
        };

        loop {
            if state.i >= state.buffer.len() {
                // finished reading segment
                if state.segment_size.is_incomplete() {
                    // read the header of the next segment and grow the buffer
                    state.segment_size =
                        Self::recv_header(&mut self.header_state, &mut self.stream).await?;

                    state
                        .buffer
                        .resize(state.buffer.len() + state.segment_size.size() as usize, 0);

                    continue;
                } else {
                    // done
                    break;
                }
            }

            let i = self.stream.read(&mut state.buffer[state.i..]).await.ok()?;

            if i == 0 {
                // unexpected EOF
                return None;
            }

            state.i += i;
        }

        // TODO: Reader?
        let message = bincode::deserialize(&state.buffer).ok();
        self.receiving_state = None;

        message
    }
}

fn random_u64() -> u64 {
    let mut bytes = [0; 8];
    getrandom::getrandom(&mut bytes).unwrap();
    u64::from_ne_bytes(bytes)
}

pub async fn handler<T, S, F>(
    address: SocketAddr,
    mut message_receiver: Receiver<T>,
    backoff: fn(u32) -> Duration,
    max_in_flight: usize,
    connect: impl Fn(TcpStream) -> F,
) where
    T: serde::Serialize + for<'de> serde::Deserialize<'de> + Send + Sync + 'static,
    S: AsyncReadExt + AsyncWriteExt,
    F: Future<Output = S>,
{
    let mut message_buffer = VecDeque::<Message<T>>::with_capacity(max_in_flight);

    let mut n_failures = Saturating(0u32);
    let mut last_connection_attempt = Instant::now();

    let mut last_sent = Wrapping(0u64);
    let mut n_in_flight = 0;
    let mut reconnect = false;

    loop {
        if !reconnect {
            last_sent = Wrapping(random_u64());
        }

        sleep_if_nonzero(backoff(n_failures.0).saturating_sub(last_connection_attempt.elapsed()))
            .await;

        last_connection_attempt = Instant::now();

        let Ok(connection) = TcpStream::connect(address).await else {
            n_failures += 1;
            continue;
        };

        let (reader, writer) = tokio::io::split(connect(connection).await);
        let mut writer = StreamWriteHalf::<T, _>::new(writer);
        let mut reader = StreamReadHalf::<T, _>::new(reader);

        let Some(_) = writer
            .send(&Message::Connect {
                reconnect,
                last_id: last_sent.0,
            })
            .await
        else {
            n_failures += 1;
            continue;
        };

        let Some(res) = reader.recv().await else {
            n_failures += 1;
            continue;
        };

        let mut last_received = match res {
            Message::Connect {
                reconnect: remote_reconnect,
                last_id,
            } if remote_reconnect == reconnect => Wrapping(last_id),
            _ => {
                n_failures += 1;
                reconnect = false;
                continue;
            }
        };

        // dbg!(last_sent);
        // dbg!(last_received);
        // dbg!(&message_buffer);
        // dbg!(n_in_flight);

        let n_not_received = last_sent - last_received;
        // dbg!(n_not_received);
        if reconnect && n_not_received.0 > n_in_flight as u64 {
            // println!("remote id is further behind than valid: {n_not_received} > {n_in_flight}");
            n_failures += 1;
            reconnect = false;
            continue;
        }

        let n_received_but_not_acknowleged = last_received.0.saturating_sub(last_sent.0);
        // dbg!(n_received_but_not_acknowleged);
        if n_received_but_not_acknowleged > n_in_flight as u64 {
            // println!("remote id is further ahead than valid: {n_received_but_not_acknowleged} > {n_in_flight}");
            n_failures += 1;
            reconnect = false;
            continue;
        }

        let n_received_but_not_acknowleged = (n_in_flight as u64).saturating_sub(n_not_received.0);
        for _ in 0..n_received_but_not_acknowleged {
            _ = message_buffer.pop_front();
            // println!("popped message from queue which was received but not acknowleged: {popped:?}");
        }

        // TODO: remove extra messages from the buffer that were received but not acknowleged
        last_sent = last_received;

        n_in_flight = 0;
        reconnect = true;

        let mut last_ping_sent = Instant::now() - PING_INTERVAL;
        let mut last_ping_recv = Instant::now();

        loop {
            n_failures.0 = 0;

            let recv_message = if message_buffer.len() >= max_in_flight {
                // there is no space in the message buffer to receive new messages
                Fuse::terminated()
            } else {
                match message_receiver.try_recv() {
                    Ok(message) => {
                        message_buffer.push_back(Message::Payload(message));
                        continue;
                    }
                    // Receive incoming messages and store them in the buffer
                    Err(TryRecvError::Empty) => message_receiver.recv().fuse(),
                    // No more messages can be received, but some are still buffered
                    Err(TryRecvError::Disconnected) if !message_buffer.is_empty() => {
                        Fuse::terminated()
                    }
                    // There are no more messages to receive
                    Err(TryRecvError::Disconnected) => {
                        println!("No more messages to send. Exiting");
                        return;
                    }
                }
            };

            // if there are messages in the message buffer that have not yet been sent.
            // Attempt to send them.
            let send = if message_buffer.len() > n_in_flight {
                async {
                    if !writer.has_partial_send() {
                        writer.start_sending(message_buffer.get(n_in_flight).unwrap());
                    }

                    writer.finish_send().await
                }
                .fuse()
            } else {
                Fuse::terminated()
            };

            select! {
                biased;
                _ = sleep_if_nonzero(PING_INTERVAL.saturating_sub(last_ping_sent.elapsed())) => {
                    writer.finish_send().await;
                    writer.send(&Message::Ping).await;
                    last_ping_sent = Instant::now();
                },
                received = reader.recv().fuse() => {
                    match received {
                        Some(Message::Ack { id }) if id == (last_received + Wrapping(1)).0 => {
                            assert!(n_in_flight > 0);
                            n_in_flight -= 1;
                            last_received += 1;
                            _ = message_buffer.pop_front();
                            // println!("popped from queue: {popped:?}");
                        }
                        Some(Message::Ping) => {
                            last_ping_recv = Instant::now();
                        }
                        _ => {
                            // println!("failed to receive an ack. Try to reconnect");
                            n_failures += 1;
                            break;
                        }
                    }
                },
                sent = send => {
                    if sent.is_some() {
                        last_sent += 1;
                        // dbg!(last_sent);
                        n_in_flight = n_in_flight.checked_add(1).unwrap();
                    } else {
                        // println!("failed to send message. Try to reconnect");
                        n_failures += 1;
                        break;
                    }
                },
                message = recv_message => {
                    if let Some(message) = message {
                        message_buffer.push_back(Message::Payload(message));
                    } else {
                        // println!("Message sender exited");
                        continue;
                    }
                },

                _ = sleep_if_nonzero(PING_TIMEOUT.saturating_sub(last_ping_recv.elapsed())) => {
                    println!("handler timeout");
                    break;
                },
            }
        }
    }
}

async fn sleep_if_nonzero(duration: Duration) {
    if !duration.is_zero() {
        tokio::time::sleep(duration).await
    }
}

pub async fn acceptor<T, S, F>(
    message_sender: Sender<T>,
    listener: TcpListener,
    accept: impl Fn(TcpStream) -> F,
) where
    T: serde::Serialize + for<'de> serde::Deserialize<'de> + Send + Sync + 'static,
    S: AsyncReadExt + AsyncWriteExt,
    F: Future<Output = S>,
{
    let mut last_recv = Wrapping(0);
    let mut reconnect = false;

    loop {
        let (stream, _remote_addr) = listener.accept().await.unwrap();
        let (reader, writer) = tokio::io::split(accept(stream).await);
        let mut writer = StreamWriteHalf::<T, _>::new(writer);
        let mut reader = StreamReadHalf::<T, _>::new(reader);

        let Some(res) = reader.recv().await else {
            // println!("failed to recv handshake");
            continue;
        };

        let res = match res {
            Message::Connect {
                reconnect: true,
                last_id: _,
            } => {
                // println!("reconnected at {last_recv}->{last_id}");

                if !reconnect {
                    // could not reesatblish connection
                    return;
                }

                writer
                    .send(&Message::Connect {
                        reconnect,
                        last_id: last_recv.0,
                    })
                    .await
            }
            Message::Connect {
                reconnect: false,
                last_id,
            } => {
                if reconnect {
                    // could not reesatblish connection
                    return;
                }

                // println!("Establishing new connection at {last_id}");
                last_recv.0 = last_id;

                writer
                    .send(&Message::Connect {
                        reconnect: false,
                        last_id,
                    })
                    .await
            }
            _ => {
                // println!("unexpected handshake packet");
                continue;
            }
        };

        if res.is_none() {
            // println!("failed to send handshake");
            continue;
        }

        reconnect = true;

        let mut last_ping_sent = Instant::now() - PING_INTERVAL;
        let mut last_ping_recv = Instant::now();

        loop {
            let message = select! {
                biased;
                _ = sleep_if_nonzero(PING_INTERVAL.saturating_sub(last_ping_sent.elapsed())) => {
                    writer.send(&Message::Ping).await;
                    last_ping_sent = Instant::now();
                    continue;
                }
                message = reader.recv() => {
                    message
                },
                _ = sleep_if_nonzero(PING_TIMEOUT.saturating_sub(last_ping_recv.elapsed())) => {
                    println!("listener timeout");
                    break;
                }
            };

            let Some(message) = message else {
                // println!("failed to recv data");
                break;
            };

            last_ping_recv = Instant::now();

            match message {
                Message::Payload(message) => {
                    let mut send = std::pin::pin!(message_sender.send(message).fuse());
                    loop {
                        select! {
                            res = &mut send => {
                                res.unwrap();
                                break;
                            },
                            _ = sleep_if_nonzero(PING_INTERVAL.saturating_sub(last_ping_sent.elapsed())) => {
                                writer.send(&Message::Ping).await;
                                last_ping_sent = Instant::now();
                            }
                        }
                    }
                    last_recv += 1;
                    if writer
                        .send(&Message::Ack { id: last_recv.0 })
                        .await
                        .is_none()
                    {
                        // println!("failed to send ack");
                        break;
                    }
                }
                Message::Ping => {}
                _ => {
                    // println!("unexpected data packet");
                    break;
                }
            }
        }
    }
}

#[cfg(test)]
mod tests {
    // #[tokio::test(flavor = "multi_thread", worker_threads = 1)]

    use std::{
        net::SocketAddr,
        pin::Pin,
        sync::Arc,
        time::{Duration, Instant},
    };

    use futures::{join, Future};
    use tokio::{
        io::{AsyncReadExt, AsyncWriteExt},
        net::{
            tcp::{OwnedReadHalf, OwnedWriteHalf},
            TcpListener, TcpStream,
        },
        sync::mpsc,
    };
    use zerocopy::{AsBytes, FromZeroes};

    use crate::{acceptor, handler, sleep_if_nonzero, Message, SegmentSize, PING_TIMEOUT, U31_MAX};

    async fn listener<T>(message_sender: tokio::sync::mpsc::Sender<T>) -> SocketAddr
    where
        T: serde::Serialize + for<'de> serde::Deserialize<'de> + Send + Sync + 'static,
    {
        let listener = TcpListener::bind(("127.0.0.1", 0)).await.unwrap();
        let addr = listener.local_addr().unwrap();
        tokio::task::spawn(acceptor(message_sender, listener, |stream| async move {
            stream
        }));
        addr
    }

    #[test]
    fn test_segment_size() {
        let mut total_size = u32::MAX as u64 / 2 + 100;
        assert!(total_size > U31_MAX as u64);

        let size = SegmentSize::from_total(total_size);
        assert!(size.is_incomplete());
        assert_eq!(size.size(), U31_MAX);
        total_size -= size.size() as u64;

        let size = SegmentSize::from_total(total_size);
        assert!(!size.is_incomplete());
        assert_eq!(size.size() as u64, total_size);
        total_size -= size.size() as u64;
        assert_eq!(total_size, 0);

        let size = SegmentSize::from_total(0);
        assert!(!size.is_incomplete());
        assert_eq!(size.size() as u64, 0);
    }

    type Data = u32;

    type HandlerFactory = Box<dyn Fn() -> Handler + Send + Sync + 'static>;
    type Handler = Box<
        dyn FnMut(
                SegmentSize,
                Message<Data>,
            ) -> Pin<Box<dyn Future<Output = bool> + Send + Sync + 'static>>
            + Send
            + Sync
            + 'static,
    >;

    async fn proxy_handler(
        mut reader: OwnedReadHalf,
        mut writer: OwnedWriteHalf,
        mut handler: Handler,
    ) -> std::io::Result<()> {
        loop {
            let mut size = SegmentSize::new_zeroed();
            let mut segments = Vec::new();
            let mut buf = Vec::new();
            let mut i = 0;
            loop {
                match reader.read_exact(size.as_bytes_mut()).await {
                    Err(err) if err.kind() == std::io::ErrorKind::UnexpectedEof => return Ok(()),
                    Err(err) => return Err(err),
                    Ok(_) => {}
                }
                buf.resize(buf.len() + size.size() as usize, 0);
                reader.read_exact(&mut buf[i..]).await.unwrap();
                segments.push((size, buf[i..].to_owned()));
                i += size.size() as usize;
                if !size.is_incomplete() {
                    break;
                }
            }

            let message: Message<Data> = bincode::deserialize(&buf).unwrap();
            if handler(size, message).await {
                break Ok(());
            }

            for (size, buf) in segments {
                writer.write_all(size.as_bytes()).await?;
                writer.write_all(&buf).await?;
            }
        }
    }

    async fn proxy(
        addr: SocketAddr,
        handler_a: HandlerFactory,
        handler_b: HandlerFactory,
    ) -> SocketAddr {
        let listener = TcpListener::bind(("127.0.0.1", 0)).await.unwrap();
        let own_addr = listener.local_addr().unwrap();

        let handler_a = Arc::new(handler_a);
        let handler_b = Arc::new(handler_b);

        tokio::task::spawn(async move {
            loop {
                let (stream, _remote_addr) = listener.accept().await.unwrap();
                let handler_a = handler_a.clone();
                let handler_b = handler_b.clone();
                tokio::task::spawn(async move {
                    let other_stream = TcpStream::connect(addr).await.unwrap();

                    let (stream_read, stream_write) = stream.into_split();
                    let (other_stream_read, other_stream_write) = other_stream.into_split();

                    let handle_a = proxy_handler(stream_read, other_stream_write, handler_a());
                    let handle_b = proxy_handler(other_stream_read, stream_write, handler_b());

                    let _ = join!(handle_a, handle_b);
                });
            }
        });

        own_addr
    }

    #[tokio::test]
    async fn good_connection() {
        let (message_sender, mut receive_handle) = mpsc::channel(1);
        let (send_handle, message_receiver) = mpsc::channel(1);

        let addr = listener::<Data>(message_sender).await;
        let handler_handle = tokio::task::spawn(handler::<Data, _, _>(
            addr,
            message_receiver,
            |_| Duration::from_secs(0),
            5,
            |stream| async move { stream },
        ));

        let messages: Arc<Vec<u32>> =
            Arc::new((0..10_000).map(|_| rand::random::<Data>()).collect());

        {
            let messages = messages.clone();
            tokio::task::spawn(async move {
                for &message in messages.iter() {
                    send_handle
                        .send(message)
                        .await
                        .expect("failed to send message");
                }
                drop(send_handle);
            });
        }

        let start = Instant::now();
        for &message in messages.iter() {
            assert_eq!(receive_handle.recv().await, Some(message));
        }
        println!("total: {:05.0?}", start.elapsed());
        drop(receive_handle);

        println!("received all messages");

        handler_handle.await.unwrap();
        println!("handler done");
    }

    // #[tokio::test(flavor = "multi_thread", worker_threads = 2)]
    #[tokio::test]
    async fn no_timeout() {
        let (message_sender, mut receive_handle) = mpsc::channel(1);
        let (send_handle, message_receiver) = mpsc::channel(1);

        let addr = listener::<Data>(message_sender).await;
        let handler_handle = tokio::task::spawn(handler::<Data, _, _>(
            addr,
            message_receiver,
            |_| Duration::from_secs(0),
            5,
            |stream| async move { stream },
        ));

        let messages: Arc<Vec<u32>> = Arc::new((0..20).map(|_| rand::random::<Data>()).collect());

        {
            let messages = messages.clone();
            tokio::task::spawn(async move {
                for &message in messages.iter() {
                    send_handle
                        .send(message)
                        .await
                        .expect("failed to send message");
                }
                drop(send_handle);
            });
        }

        for (i, &message) in messages.iter().enumerate() {
            let start = Instant::now();
            assert_eq!(receive_handle.recv().await, Some(message));
            println!("{i:4}: {:05.0?}", start.elapsed());

            if i % 10 == 0 {
                sleep_if_nonzero(2 * PING_TIMEOUT).await;
            }
        }
        drop(receive_handle);

        println!("received all messages");

        handler_handle.await.unwrap();
        println!("handler done");
    }

    #[tokio::test]
    async fn very_bad_connection() {
        // let counter_a = Arc::new(AtomicUsize::new(0));
        let handler_a = Box::new(move || {
            // let counter = counter_a.clone();
            Box::new(move |_size, _message_a| {
                // print!("A: {message_a:?}");
                Box::pin(async move {
                    if rand::random::<f32>() < 0.1 {
                        sleep_if_nonzero(PING_TIMEOUT.mul_f32(1.1)).await;
                    }

                    rand::random::<f32>() < 0.1
                }) as Pin<Box<dyn Future<Output = bool> + Send + Sync + 'static>>
            }) as Handler
        }) as HandlerFactory;

        // let counter_b = Arc::new(AtomicUsize::new(0));
        let handler_b = Box::new(move || {
            // let counter = counter_b.clone();
            Box::new(move |_size, _message_b| {
                // print!("B: {message_b:?}");
                Box::pin(async move {
                    if rand::random::<f32>() < 0.1 {
                        sleep_if_nonzero(PING_TIMEOUT.mul_f32(1.1)).await;
                    }

                    rand::random::<f32>() < 0.1
                }) as Pin<Box<dyn Future<Output = bool> + Send + Sync + 'static>>
            }) as Handler
        }) as HandlerFactory;

        let (message_sender, mut receive_handle) = mpsc::channel(1);
        let addr = listener::<Data>(message_sender).await;
        let proxy = proxy(addr, handler_a, handler_b).await;

        let (send_handle, message_receiver) = mpsc::channel(1);

        let handler_handle = tokio::task::spawn(handler::<Data, _, _>(
            proxy,
            message_receiver,
            |_| Duration::from_secs(0),
            5,
            |stream| async move { stream },
        ));

        let messages: Arc<Vec<u32>> = Arc::new((0..50).map(|_| rand::random::<Data>()).collect());

        {
            let messages = messages.clone();
            tokio::task::spawn(async move {
                for &message in messages.iter() {
                    send_handle
                        .send(message)
                        .await
                        .expect("failed to send message");
                }
                drop(send_handle);
            });
        }

        for (i, &message) in messages.iter().enumerate() {
            let start = Instant::now();
            assert_eq!(receive_handle.recv().await, Some(message));
            println!("{i:4}: {:05.0?}", start.elapsed());
        }
        drop(receive_handle);

        println!("received all messages");

        handler_handle.await.unwrap();
        println!("handler done");
    }
}
